/**
 * See klass on mõeldud suhtluseks Kasutajaliidese ja Expression vahel
 */

package ee.ut.cs.gitlab.maikov;

import ee.ut.cs.gitlab.maikov.exceptions.*;
import ee.ut.cs.gitlab.maikov.expressions.Expression;
import ee.ut.cs.gitlab.maikov.expressions.MatrixValue;
import ee.ut.cs.gitlab.maikov.expressions.Value;
import ee.ut.cs.gitlab.maikov.matrix.Matrix;
import ee.ut.cs.gitlab.maikov.matrix.SquareMatrix;

import java.util.HashMap;
import java.util.regex.Pattern;

public class Vahendaja {

    // Hoiab kõiki muutujaid, mis lisatakse
    private HashMap<String, Value> muutujad;

    public Vahendaja() {
        muutujad = new HashMap<String, Value>();
    }


    public Value leiaMuutuja(String nimi) {
        return muutujad.get(nimi);
    }


    /**
     * Kas nimi koosneb vaid tähtedest
     */
    public static boolean tähedP(String nimi) {
        return Pattern.matches("[a-zA-ZõäöõÕÄÖÜ]*", nimi);
    }

    /**
     * Kas on juba selle nimega maatriks?
     */
    public boolean olemasP(String maatriksi_nimi) {
        return (muutujad.containsKey(maatriksi_nimi));
    }

    /**
     * Kas muutuja nimi sobib
     * Kui nimi sobib, 1
     * Kui selle nimega muutuja juba olemas, 0
     * Kui nimes ei ole ainult tähed, -1
     */
    public int sobivNimiP(String nimi) {
        if (nimi.equals("T")) {
            return -2;
        } else if (!tähedP(nimi)) {
            return -1;
        } else if (olemasP(nimi)) {
            return 0;
        } else {
            return 1;
        }
    }

    /**
     * Loob maatriksi objekti
     */
    public void lisaMaatriks(String nimi, int[] mõõtmed, double[] maatriks) {
        if (this.sobivNimiP(nimi) != 1) {
            System.out.println(nimi + " on juba olemas või sisaldab tähemärke, mis ei ole tähed");
            return;
        }
        muutujad.put(nimi,
                new MatrixValue((mõõtmed[0] == mõõtmed[1])
                        ? new SquareMatrix(mõõtmed[1], maatriks)
                        : new Matrix(mõõtmed[0], mõõtmed[1], maatriks)));
    }


    /**
     * Tagastab sõne kõikide muutujate väärtustega
     */
    public String väärtusedList() {
        String väljund = "";
        for (String nimi : muutujad.keySet()) {
            väljund += nimi;
            väljund += System.lineSeparator();
            väljund += muutujad.get(nimi).toString();
            väljund += System.lineSeparator();
        }
        return väljund;
    }


    /**
     * Parsib ohutult sõne
     * Püüab vead kinni
     */
    public Value parse_wrap(String võrrand) {
        try {

            Expression e = Expression.parse(võrrand, muutujad);
            return (Value) e.compute();
        } catch (TundmatuMuutuja v) {
            System.out.println("Tundmatu Muutuja: " + v.getMessage());
            return null;
        } catch (InvalidOperandException o) {
            System.out.println("InvalidOperandException:");
            System.out.println(o.getMessage());
            return null;
        } catch (TõlgendamatuSisend s) {
            System.out.println("Tõlgendamatu sisend:");
            System.out.println(s.getMessage());
            return null;
        } catch (InvalidReductionException r) {
            System.out.println("Tehte sooritamisel tekkis viga.");
            return null;
        } catch (IllegalArgumentException a) {
            System.out.println("IllegalArgumentException:");
            System.out.println(a.getMessage());
            return null;
        } catch (InvalidMatrixException e) {
            System.out.println("Invaliidne maatriks:");
            System.out.println(e.getMessage());
            return null;
        }
    }

    /**
     * Annab muutujale <nimi> väärtuse, mille saav <võrrand> lugedes
     * Kui <nimi> ei sobi muutuja nimeks, teatab veast.
     */
    public void väärtusta(String nimi, String võrrand) {
        if (sobivNimiP(nimi) == 1) {
            Value n = parse_wrap(võrrand);
            if (n != null) {
                muutujad.put(nimi, n);
                System.out.println(n);
            }
        } else {
            System.out.println("EBASOBIV NIMI: " + nimi);
        }
    }

}
