package ee.ut.cs.gitlab.maikov.expressions;

import ee.ut.cs.gitlab.maikov.exceptions.InvalidOperandException;
import ee.ut.cs.gitlab.maikov.exceptions.TundmatuMuutuja;
import ee.ut.cs.gitlab.maikov.exceptions.TõlgendamatuSisend;
import ee.ut.cs.gitlab.maikov.matrix.Matrix;

import java.util.ArrayList;
import java.util.Map;
import java.util.regex.Pattern;

public abstract class Expression {
    public abstract Expression compute();

    // Eraldab iga väärtuse, sulu ja tehtemärgi eraldi sõnesse
    // Autor: Harald Jõgi
    private static ArrayList<String> lex(String võrrand) {
        ArrayList<String> märgid = new ArrayList<>();
        // Pinu kõrgus

        // Parasjagu loetav number või nimi
        String kogum = "";

        for (int i = 0; i < võrrand.length(); i++) {
            Character a = võrrand.charAt(i);

            if (a.equals(' ') || a.equals(',')) { // Tühikud ja komad
                if (!kogum.equals("")) {
                    märgid.add(kogum);
                    kogum = "";
                }
            } else if (Character.isLetter(a)) { // Tähed
                kogum += a;
            } else if (Character.isDigit(a) || a.equals('.')) { // Numbrid
                kogum += a;
            } else { // Ülejäänud
                if (!kogum.equals("")) {
                    märgid.add(kogum);
                    kogum = "";
                }
                märgid.add(a.toString());
            }
        }
        if (!kogum.equals("")) {
            märgid.add(kogum);
            kogum = "";
        }

        return märgid;
    }

    // Teeb nii palju tehteid paremalt poolt ära, kui ohutult võimalik
    // Autor: Harald Jõgi
    private static void fold(ArrayList<String> tehted,
                             ArrayList<Expression> vasak)
            throws InvalidOperandException {

        //System.out.println("\nTEHTED: "+tehted+"\nVASAK: "+vasak);



        int T_index = tehted.size() - 2; // Ignoreerib just lisatut
        int V_index = vasak.size() - 1;

        // Just lisatu
        String uusim = tehted.get(tehted.size() - 1);
        // Kuna läheme järgmise suluni, pole seda vaja
        if (uusim.equals(")") || uusim.equals("|")) {
            tehted.remove(tehted.size() - 1);
        }

        for (; T_index >= 0; T_index--) {
            String tehe = tehted.get(T_index);
            // Ühe muutuja tehted
            if (tehe.equals("(")) {  // Lõpetab, kui suluni jõuab
                if (uusim.equals(")")) {
                    tehted.remove(T_index);
                }
                break;
            } else if (tehe.equals("|")) {  // Lõpetab, kui suluni jõuab
                if (uusim.equals("|")) {
                    tehted.remove(T_index);
                    vasak.set(V_index, new AbsoluteValue(vasak.get(V_index)));
                }
                break;
            } else if (tehe.equals("vastandarv")){ // Vastandarv
                // Korrutab '-1'-ga
                vasak.set(V_index, new Multiplication(new Number(-1),
                        vasak.get(V_index)));
                tehted.remove(T_index);

            } // Kahe muutuja tehted
            else if (tehe.equals("^")) { // Pöörd- või transponeeritud maatriks
                if (V_index <= 0){
                    throw new TõlgendamatuSisend("Tehe "+tehe+" vajab kahte muutujat!");
                }
                
                Expression baas = vasak.get(V_index - 1);
		
		vasak.set(V_index-1,
			  new Exponentiation(baas,vasak.get(V_index)));
		
                
                tehted.remove(T_index);
                vasak.remove(V_index--);
            } else if ("+-*/".contains(tehe)) {
                if (V_index <= 0){
                    throw new TõlgendamatuSisend("Tehe "+tehe+" vajab kahte muutujat!");
                }

                Expression V = vasak.get(V_index - 1);
                Expression P = vasak.get(V_index);
                switch (tehe) {
                    case "+":
                        vasak.set(V_index - 1, new Addition(V, P));
                        break;
                    case "-":
                        vasak.set(V_index - 1, new Subtraction(V, P));
                        break;
                    case "*":
                        vasak.set(V_index - 1, new Multiplication(V, P));
                        break;
                    case "/":
                        vasak.set(V_index - 1, new Division(V, P));
                        break;
                }
                vasak.remove(V_index--);
                tehted.remove(T_index);
            }
        }
        //System.out.println("\nPÄrast\nTEHTED: "+tehted+"\nVASAK: "+vasak+"\n\n");

    }


    // Loeb sõnest tehted
    // Autor: Harald Jõgi
    public static Expression parse(String võrrand, Map<String, Value> liikmed)
            throws TundmatuMuutuja, InvalidOperandException, TõlgendamatuSisend{
        // võrrand, jagatud vähimateks osadeks
        ArrayList<String> lexed = lex(võrrand);
        // Tehtemärgid ja sulud järjekorras, milles nendega kohtume
        // Kui tehte mõlemad pooled olemas, teeme Expression'iks
        ArrayList<String> tehted = new ArrayList<>();
        // Väärtused järjekorras, milles neid näeme
        // Kui tehte mõlemad pooled olemas, teeme Expression'iks ja asendame
        // sellega esimese väärtuse
        ArrayList<Expression> vasak = new ArrayList<>();

        // Eelmine lekseem
        String eelmine = "";
        // Kas on parasjagu püstkriips
        boolean püst = false;
        // Kas proovida tehteid sooritada
        boolean f = false;



        for (int l = 0; l < lexed.size(); l++) {
            String lekseem = lexed.get(l);

            if ("+)".contains(lekseem)) { // Tehtemärgid + ja )
                tehted.add(lekseem);
                f = true;
            } else if (lekseem.equals("-")) {
                if ("+-*/(|^".contains(eelmine) || eelmine.equals("")){ // '-' kui arvu negatiivne variant jne
                    tehted.add("vastandarv");
                } else { // '-' kui tehe
                    tehted.add(lekseem);
                    f = true;
                }

            } else if ("*/(".contains(lekseem)) { // Tehtemärgid *,/ ning (
                if (tehted.size() != 0
                        && tehted.get(tehted.size() - 1).equals("^")) {
                    f = true;
                }
                tehted.add(lekseem);
            } else if (lekseem.equals("^")) { // ^
                tehted.add(lekseem);
            } else if (lekseem.equals("|")) { // Determinant / Absoluutväärtus
                püst = !püst;
                tehted.add(lekseem);
                if (!püst) {
                    f = true;
                    lekseem = "_"; // Et '-' tõlgendataks absoluutväärtuse järel õigesti
                }
            } else if (lekseem.equals("{")) { // Maatriks {}

                tehted.add(lekseem);
                int rows = -1;
                ArrayList<Double> elements = new ArrayList<>();
                // Läheb edasi, kuni loogelised sulud on tasakaalus
                while (true) {
                    lekseem = lexed.get(++l);

                    if (lekseem.equals("{")) {
                        tehted.add(lekseem);
                    } else if (lekseem.equals("}")) {
                        rows += 1;
                        tehted.remove(tehted.size() - 1);
                        if (tehted.size() == 0
                                || !tehted.get(tehted.size() - 1).equals("{")) {
                            if (rows == 0){
                                throw new TõlgendamatuSisend("Maatriksil peab olema vähemalt üks rida");
                            }
                            double[] elts = new double[elements.size()];
                            for (int e = 0; e < elts.length; e++) {
                                elts[e] = elements.get(e);
                            }

                            Matrix m = new Matrix(rows,
                                    elts.length / rows,
                                    elts);
                            vasak.add(new MatrixValue(m));
                            break;

                        }
                    } else {
                        elements.add(Double.parseDouble(lekseem));
                    }
                }


            } else if (Pattern.matches("-?\\d*[.]?\\d*", lekseem)) { // Arvud
                vasak.add(new Number(Double.parseDouble(lekseem)));
                // SÜMBOLID
            } else if (Pattern.matches("[a-zA-ZõäöõÕÄÖÜ]*", lekseem)) {
                if (tehted.size() > 0
                        && lekseem.equals("T") // Transponeerimine
                        && tehted.get(tehted.size() - 1).equals("^")) {
                    vasak.add(new Number(Double.NaN)); // Sümboliseerib transp.
                } else if (liikmed.containsKey(lekseem)) {
                    vasak.add(liikmed.get(lekseem).clone());
                } else {

                    throw new TundmatuMuutuja(lekseem);
                }
            } else {
                throw new TõlgendamatuSisend("TUNDMATU LEKSEEM: " + lekseem);
            }
            eelmine = lekseem;
            // Läheb läbi kogutud tehete nii palju, kui võimalik
            if (f == true) {
                f = false;

                fold(tehted, vasak);
            }

        }

        tehted.add("NOP"); // Kuna ei vaadata viimast tehet
        fold(tehted, vasak);


        // Kui ei ole lõpuks täpselt üks väärtus ja null tehet
        if (vasak.size() != 1 || tehted.size() != 1){
            throw new TõlgendamatuSisend("Sisendist ei saanud võrrandit");
        }
        return vasak.get(0);
    }
}
