package ee.ut.cs.gitlab.maikov.expressions;

import ee.ut.cs.gitlab.maikov.exceptions.InvalidMatrixException;
import ee.ut.cs.gitlab.maikov.exceptions.InvalidOperandException;
import ee.ut.cs.gitlab.maikov.matrix.InverseMatrix;
import ee.ut.cs.gitlab.maikov.matrix.Matrix;
import ee.ut.cs.gitlab.maikov.matrix.SquareMatrix;

public class MatrixValue extends Value {

    private final Matrix matrix;

    public MatrixValue(Matrix matrix) {
        if (matrix == null) {
            throw new NullPointerException();
        }
        this.matrix = matrix;
    }

    public Matrix getMatrix() {
        return matrix;
    }

    @Override
    public Value addition(Value other) {
        if (other instanceof MatrixValue otherMatrixValue) {
            Matrix otherMatrix = otherMatrixValue.getMatrix();
            this.matrix.add(otherMatrix);
            return new MatrixValue(this.matrix);
        } else {
            throw new InvalidOperandException("Ei saa liita maatriksit numbriga või millegi muuga.");
        }
    }

    @Override
    public Value subtraction(Value other) {
        if (other instanceof MatrixValue otherMatrixValue) {
            Matrix otherMatrix = otherMatrixValue.getMatrix();
            otherMatrix.multiply(-1);
            this.matrix.add(otherMatrix);
            return new MatrixValue(this.matrix);
        } else {
            throw new InvalidOperandException("Ei saa lahutada maatriksit numbriga või millegi muuga.");
        }
    }

    @Override
    public Value multiplication(Value other) {
        if (other instanceof Number otherNumber) {
            this.matrix.multiply(otherNumber.getNumber());
            return new MatrixValue(this.matrix);
        } else if (other instanceof MatrixValue otherMatrixValue) {
            Matrix otherMatrix = otherMatrixValue.getMatrix();
            this.matrix.multiply(otherMatrix);
            return new MatrixValue(this.matrix);
        } else {
            throw new InvalidOperandException("Tundmatu väärtuse tüüp!");
        }
    }

    @Override
    public Value division(Value other) {
        throw new InvalidOperandException();
    }

    @Override
    public Value absoluteValue() {
        if (matrix instanceof SquareMatrix squareMatrix) {
            return new Number(squareMatrix.determinant());
        } else if (matrix.getRows() == matrix.getCols()) {
            SquareMatrix squareMatrix = new SquareMatrix(matrix.getRows(), matrix.getElements());
            return new Number(squareMatrix.determinant());
        } else {
            throw new InvalidMatrixException("Tegemist ei ole ruutmaatriksiga!");
        }
    }

    private SquareMatrix getSquareMatrix() {
        if (this.getMatrix() instanceof SquareMatrix squareMatrix) {
            return squareMatrix;
        } else if (this.getMatrix().getRows() == this.getMatrix().getCols()) {
            return new SquareMatrix(this.getMatrix().getRows(), this.getMatrix().getElements());
        } else {
            throw new InvalidMatrixException("Ei ole ruutmaatriks.");
        }
    }

    @Override
    public Value exponentiation(Value other) {
        if (other instanceof Number otherNumber) {
            if (Double.isFinite(otherNumber.getNumber())) {
                SquareMatrix squareMatrix = getSquareMatrix();
                int n = (int) otherNumber.getNumber();
                if (n == 0) {
                    squareMatrix.identity();
                    return new MatrixValue(squareMatrix);
                } else if (n < 0) {
                    return new MatrixValue(InverseMatrix.compute(squareMatrix)).exponentiation(new Number(-n));
                } else {
                    Matrix result = Matrix.clone(this.getMatrix());
                    for (int i = 1; i < n; i++) {
                        result.multiply(this.getMatrix());
                    }
                    return new MatrixValue(result);
                }
            } else {
                return new MatrixValue(this.matrix.transpose());
            }

        } else {
            throw new InvalidOperandException("Ei saa astendada maatriksit maatriksiga või millegi muuga.");
        }
    }

    @Override
    public String toString() {
        return matrix.toString();
    }

    @Override
    public Value clone() {
        return new MatrixValue(Matrix.clone(this.matrix));
    }
}
