package ee.ut.cs.gitlab.maikov.expressions;

import ee.ut.cs.gitlab.maikov.exceptions.InvalidOperandException;
import ee.ut.cs.gitlab.maikov.matrix.Matrix;

public class Number extends Value {

    private final double number;

    public Number(double number) {
        this.number = number;
    }

    public double getNumber() {
        return number;
    }

    @Override
    public Value addition(Value other) {
        if (other instanceof Number otherNumber) {
            return new Number(this.getNumber() + otherNumber.getNumber());
        } else {
            throw new InvalidOperandException("Ei saa liita numbrit ja maatriksit või midagi muud.");
        }
    }

    @Override
    public Value subtraction(Value other) {
        if (other instanceof Number otherNumber) {
            return new Number(this.getNumber() - otherNumber.getNumber());
        } else {
            throw new InvalidOperandException("Ei saa lahutada numbrit ja maatriksit või midagi muud");
        }
    }

    @Override
    public Value multiplication(Value other) {
        if (other instanceof Number otherNumber) {
            return new Number(this.getNumber() * otherNumber.getNumber());
        } else if (other instanceof MatrixValue otherMatrixValue) {
            Matrix otherMatrix = otherMatrixValue.getMatrix();
            otherMatrix.multiply(this.getNumber());
            return new MatrixValue(otherMatrix);
        } else {
            throw new InvalidOperandException("Mingi tundmatu väärtuse tüüp.");
        }
    }

    @Override
    public Value division(Value other) {
        if (other instanceof Number otherNumber) {
            return new Number(this.getNumber() / otherNumber.getNumber());
        } else {
            throw new InvalidOperandException("Ei saa jagada numbrit maatriksiga või millegi muuga.");
        }
    }

    @Override
    public Value absoluteValue() {
        return new Number(Math.abs(this.getNumber()));
    }

    @Override
    public Value exponentiation(Value other) {
        if (other instanceof Number otherNumber) {
            return new Number(Math.pow(this.getNumber(), otherNumber.getNumber()));
        } else {
            throw new InvalidOperandException("Ei saa astendata numbrit maatriksiga või millegi muuga.");
        }
    }

    @Override
    public String toString() {
        return Double.toString(number);
    }

    @Override
    public Value clone() {
        return new Number(this.number);
    }
}
