package ee.ut.cs.gitlab.maikov.matrix;

public class InverseMatrix extends Matrix {

    public InverseMatrix(SquareMatrix matrix) {
        super(matrix.cols, matrix.cols*2);

        if (matrix.determinant() == 0) {
            throw new IllegalArgumentException("Matrix determinant is zero");
        }

        for (int row = 0; row < matrix.cols; row++) {
            for (int col = 0; col < matrix.cols; col++) {
                this.set(row, col, matrix.get(row, col));
            }
        }

        int rowIndex = 0;
        for (int i = matrix.cols; i < cols; i++) {
            this.set(rowIndex++, i, 1.0);
        }
    }

    private void divideRowBy(int row, double factor) {
        for (int col = 0; col < this.cols; col++) {
            this.elements[this.getIndex(row, col)] /= factor;
        }
    }

    private void subtractRow(int row, double[] rows) {
        for (int col = 0; col < this.cols; col++) {
            this.elements[this.getIndex(row, col)] -= rows[col];
        }
    }

    private double[] getRowMultiple(int row, double factor) {
        double[] result = new double[this.cols];
        for (int col = 0; col < this.cols; col++) {
            result[col] = this.get(row, col) * factor;
        }
        return result;
    }

    public void gaussJordonMethod() {

        for (int i = 0; i < this.rows; i++) {
            double factor = this.get(i, i);

            divideRowBy(i, factor);

            for (int j = 0; j < this.rows; j++) {
                if (j == i) continue;

                double otherFactor = this.get(j, i);
                double[] rowMultiple = this.getRowMultiple(i, otherFactor);
                subtractRow(j, rowMultiple);
            }
        }
    }

    public SquareMatrix getInverseMatrix() {
        SquareMatrix result = new SquareMatrix(rows);
        for (int row = 0; row < rows; row++) {
            for (int col = 0; col < this.cols/2; col++) {
                result.set(row, col, this.get(row, col + this.cols/2));
            }
        }
        return result;
    }

    public static SquareMatrix compute(SquareMatrix matrix) {
        InverseMatrix inverseMatrix = new InverseMatrix(matrix);
        inverseMatrix.gaussJordonMethod();
        return inverseMatrix.getInverseMatrix();
    }
}
