package ee.ut.cs.gitlab.maikov.matrix;

public class SquareMatrix extends Matrix {

    public SquareMatrix(int n) {
        super(n, n);
    }

    public SquareMatrix(int n, double[] elements) {
        super(n, n, elements);
    }

    public void identity() {
        for (int i = 0; i < this.cols; i++) {
            this.set(i, i, 1.0);
        }
    }

    public double getDiagonalSum() {
        double sum = 0.0;
        for (int i = 0; i < this.cols; i++) {
            sum += this.get(i, i);
        }
        return sum;
    }

    public double getDiagonalProduct() {
        double sum = 1.0;
        for (int i = 0; i < this.cols; i++) {
            sum *= this.get(i, i);
        }
        return sum;
    }

    private boolean isTriangularAbove() {
        for (int row = 0; row < this.rows; row++) {
            for (int col = row + 1; col < this.cols; col++) {
                if (this.get(row, col) != 0.0) {
                    return false;
                }
            }
        }
        return true;
    }

    private boolean isTriangularBelow() {
        for (int row = 0; row < this.rows; row++) {
            for (int col = row + 1; col < this.cols; col++) {
                if (this.get(col, row) != 0.0) {
                    return false;
                }
            }
        }
        return true;
    }

    public boolean isTriangular() {
        return isTriangularAbove() || isTriangularBelow();
    }

    public SquareMatrix minorMatrix(int row, int col) {
        SquareMatrix result = new SquareMatrix(this.rows - 1);

        int rowOffset = 0;
        for (int rowI = 0; rowI < this.rows; rowI++) {
            if (rowI == row) {
                rowOffset = -1;
                continue;
            }

            int colOffset = 0;
            for (int colI = 0; colI < this.cols; colI++) {

                if (colI == col) {
                    colOffset = -1;
                    continue;
                }

                result.set(rowI + rowOffset, colI + colOffset, this.get(rowI, colI));
            }
        }

        return result;
    }

    public double determinant() {
        return determinant(this);
    }

    private static double determinant(SquareMatrix matrix) {
        if (matrix.rows == 1) {
            return matrix.elements[0];
        } else {
            double sum = 0.0;
            int row = 0;
            for (int col = 0; col < matrix.cols; col++) {
                int sign = (col % 2 == 0) ? 1 : -1;
                sum += sign * matrix.get(row, col) * determinant(matrix.minorMatrix(row, col));
            }
            return sum;
        }
    }
}
