import ee.ut.cs.gitlab.maikov.expressions.*;
import ee.ut.cs.gitlab.maikov.expressions.Number;
import ee.ut.cs.gitlab.maikov.matrix.Matrix;
import org.junit.jupiter.api.Test;
import static org.junit.jupiter.api.Assertions.*;

// TODO: Write more tests
public class ExpressionTest {

    private static final double DELTA = 1e-15;

    @Test
    public void testAdditionOnNumbers0() {
        Expression expr = new Addition(new Number(4), new Number(3));
        Expression result = expr.compute();
        assertInstanceOf(Number.class, result);

        Number value = (Number) result;
        assertEquals(7, value.getNumber());
    }

    @Test
    public void testAdditionOnNumbers1() {
        Expression expr = new Addition(
                new Addition(new Number(4), new Number(5)),
                new Addition(new Number(3), new Number(2)));
        Expression result = expr.compute();
        assertInstanceOf(Number.class, result);

        Number value = (Number) result;
        assertEquals(14, value.getNumber(), DELTA);
    }

    @Test
    public void testAdditionOnMatrices0() {
        Matrix a = new Matrix(3, 4, new double[] {
                1, 2, 3, 4,
                5, 6, 7, 8,
                9, 10, 11, 12,
        });
        Matrix b = new Matrix(3, 4, new double[] {
                4, 5, 9, 1,
                -5, 1, 0, 7,
                9, 4, 7, 2,
        });
        Expression expr = new Addition(new MatrixValue(a), new MatrixValue(b));
        Expression result = expr.compute();
        assertInstanceOf(MatrixValue.class, result);
        MatrixValue value = (MatrixValue) result;
        Matrix resultMatrix = value.getMatrix();
        assertEquals(3, resultMatrix.getRows());
        assertEquals(4, resultMatrix.getCols());
        assertArrayEquals(new double[]{
                5, 7, 12, 5,
                0, 7, 7, 15,
                18, 14, 18, 14
        }, resultMatrix.getElements(), DELTA);
    }

    @Test
    public void testSubtractionOnNumbers0() {

        Expression expr = new Subtraction(new Number(4), new Number(3));
        Expression result = expr.compute();
        assertInstanceOf(Number.class, result);

        Number value = (Number) result;
        assertEquals(1, value.getNumber(), DELTA);
    }

    @Test
    public void testSubtractionOnNumbers1() {

        // (4 - 1) - (7 - 2) = -2
        Expression expr = new Subtraction(
                new Subtraction(new Number(4), new Number(1)),
                new Subtraction(new Number(7), new Number(2)));
        Expression result = expr.compute();
        assertInstanceOf(Number.class, result);

        Number value = (Number) result;
        assertEquals(-2, value.getNumber(), DELTA);
    }

    @Test
    public void testSubtractionOnNumbers2() {

        // 4 - 1 - 7 - 2 = -6
        Expression expr =
                new Subtraction(
                        new Subtraction(
                                new Subtraction(new Number(4), new Number(1)), new Number(7)), new Number(2)
                );
        Expression result = expr.compute();
        assertInstanceOf(Number.class, result);

        Number value = (Number) result;
        assertEquals(-6, value.getNumber(), DELTA);
    }

    @Test
    public void testMultiplicationOnNumbers0() {
        Expression expr = new Multiplication(new Number(4), new Number(3));
        Expression result = expr.compute();

        assertInstanceOf(Number.class, result);
        Number value = (Number) result;
        assertEquals(12, value.getNumber(), DELTA);
    }

    @Test
    public void testMultiplicationOnNumbers1() {
        Expression expr = new Multiplication(new Multiplication(new Number(5), new Number(3)), new Number(2));
        Expression result = expr.compute();

        assertInstanceOf(Number.class, result);
        Number value = (Number) result;
        assertEquals(30, value.getNumber(), DELTA);
    }

    @Test
    public void testDivisionOnNumbers0() {
        Expression expr = new Division(new Number(5), new Number(2));
        Expression result = expr.compute();

        assertInstanceOf(Number.class, result);
        Number value = (Number) result;
        assertEquals(2.5, value.getNumber(), DELTA);
    }

    @Test
    public void testDivisonOnNumbers1() {
        // (5/4)/2 = 0,625
        Expression expr = new Division(new Division(new Number(5), new Number(4)), new Number(2));
        Expression result = expr.compute();

        assertInstanceOf(Number.class, result);
        Number value = (Number) result;
        assertEquals(0.625, value.getNumber(), DELTA);
    }

    @Test
    public void testMatrixTranspose0() {
        Matrix a = new Matrix(3, 4, new double[] {
                1, 2, 3, 4,
                5, 6, 7, 8,
                9, 10, 11, 12,
        });
        Expression expr = new Exponentiation(new MatrixValue(a), new Number(Double.NaN));
        Expression result = expr.compute();
        assertInstanceOf(MatrixValue.class, result);
        MatrixValue value = (MatrixValue) result;
        Matrix resultMatrix = value.getMatrix();
        assertEquals(4, resultMatrix.getRows());
        assertEquals(3, resultMatrix.getCols());
        assertArrayEquals(new double[]{
                1, 5, 9,
                2, 6, 10,
                3, 7, 11,
                4, 8, 12,
        }, resultMatrix.getElements(), DELTA);
    }

    @Test
    public void testMatrixTranspose1() {
        Matrix a = new Matrix(3, 4, new double[] {
                1, 2, 3, 4,
                5, 6, 7, 8,
                9, 10, 11, 12,
        });
        Matrix b = new Matrix(4, 3, new double[] {
                4, 2, 1,
                7, 5, 3,
                0, 8, 6,
                1, 19, 12,
        });
        Expression expr = new Exponentiation(new Multiplication(new MatrixValue(a), new MatrixValue(b)), new Number(Double.NaN));
        Expression result = expr.compute();
        assertInstanceOf(MatrixValue.class, result);
        MatrixValue value = (MatrixValue) result;
        Matrix resultMatrix = value.getMatrix();
        assertEquals(3, resultMatrix.getRows());
        assertEquals(3, resultMatrix.getCols());
        assertArrayEquals(new double[]{
                22, 70, 118,
                112, 248, 384,
                73, 161, 249
        }, resultMatrix.getElements(), DELTA);
    }

    @Test
    public void testMatrixDeterminant0() {
        Matrix a = new Matrix(3, 4, new double[] {
                1, 2, 3, 4,
                5, 6, 7, 8,
                9, 10, 11, 12,
        });
        Matrix b = new Matrix(4, 3, new double[] {
                4, 2, 1,
                7, 5, 3,
                0, 8, 6,
                1, 19, 12,
        });
        Expression expr = new AbsoluteValue(new Multiplication(new MatrixValue(a), new MatrixValue(b)));
        Expression result = expr.compute();
        assertInstanceOf(Number.class, result);
        Number value = (Number) result;
        double det = value.getNumber();
        assertEquals(0, det, DELTA);
    }

    @Test
    public void testMatrixInverse0() {
        Matrix a = new Matrix(3, 3, new double[] {
                1, 2, 3,
                0, 5, 6,
                7, 8, 9,
        });
        // A * A^-1 = E
        Expression expr = new Multiplication(new MatrixValue(a), new Exponentiation(new MatrixValue(a), new Number(-1)));
        Expression result = expr.compute();
        assertInstanceOf(MatrixValue.class, result);
        MatrixValue value = (MatrixValue) result;
        Matrix resultMatrix = value.getMatrix();
        assertEquals(3, resultMatrix.getRows());
        assertEquals(3, resultMatrix.getCols());
        assertArrayEquals(new double[]{
                1, 0, 0,
                0, 1, 0,
                0, 0, 1
        }, resultMatrix.getElements(), DELTA);
    }

    @Test
    public void testNumberExponentiation() {
        Expression expr = new Exponentiation(new Number(4), new Number(2));
        Expression result = expr.compute();
        assertInstanceOf(Number.class, result);
        double value = ((Number) result).getNumber();
        assertEquals(16, value, DELTA);
    }

    @Test
    public void testMatrixExponentiation() {
        Matrix matrix = new Matrix(3, 3, new double[]{
                4, 3, -1,
                0, 3, 7,
                9, 1, 2
        });
        Expression expr = new Exponentiation(new MatrixValue(matrix), new Number(2));
        Expression result = expr.compute();
        assertInstanceOf(MatrixValue.class, result);
        MatrixValue value = (MatrixValue) result;
        Matrix resultMatrix = value.getMatrix();
        assertEquals(3, resultMatrix.getRows());
        assertEquals(3, resultMatrix.getCols());
        assertArrayEquals(new double[]{
                7, 20, 15,
                63, 16, 35,
                54, 32, 2
        }, resultMatrix.getElements(), DELTA);
    }
}
