import ee.ut.cs.gitlab.maikov.matrix.Matrix;
import org.junit.jupiter.api.Test;
import static org.junit.jupiter.api.Assertions.*;

public class MatrixTest {

    private static final double DELTA = 1e-15;

    @Test
    void testAddition() {
        Matrix a = new Matrix(3, 3, new double[]{1, 2, 3, 4, 5, 6, 7, 8, 9});
        Matrix b = new Matrix(3, 3, new double[]{1, 2, 3, 4, 5, 6, 7, 8, 9});
        a.add(b);

        assertArrayEquals(new double[]{
                2, 4, 6,
                8, 10, 12,
                14, 16, 18
        }, a.getElements(), DELTA);
    }

    @Test
    void testMultiplication() {
        Matrix a = new Matrix(4, 3, new double[]{1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4});
        Matrix b = new Matrix(3, 4, new double[]{1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3});
        a.multiply(b);

        assertEquals(4, a.getRows());
        assertEquals(4, a.getCols());
        assertArrayEquals(new double[]{
                6, 6, 6, 6,
                12, 12, 12, 12,
                18, 18, 18, 18,
                24, 24, 24, 24
        }, a.getElements(), DELTA);
    }

    @Test
    void testMultiplicationByConstant() {
        Matrix a = new Matrix(4, 3, new double[]{1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4});
        a.multiply(3.0);

        assertEquals(4, a.getRows());
        assertEquals(3, a.getCols());
        assertArrayEquals(new double[]{
                3, 3, 3,
                6, 6, 6,
                9, 9, 9,
                12, 12, 12
        }, a.getElements(), DELTA);
    }

    @Test
    void throwsExceptionAddition() {
        Matrix a = new Matrix(3, 3, new double[]{1, 2, 3, 4, 5, 6, 7, 8, 9});
        Matrix b = new Matrix(4, 3, new double[]{1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12});

        assertThrows(IllegalArgumentException.class, () -> a.add(b));
    }

    @Test
    void throwsExceptionMultiplication() {
        Matrix a = new Matrix(3, 3, new double[]{1, 2, 3, 4, 5, 6, 7, 8, 9});
        Matrix b = new Matrix(4, 3, new double[]{1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12});

        assertThrows(IllegalArgumentException.class, () -> a.multiply(b));
    }

    @Test
    void throwsExceptionInvalidRows() {
        assertThrows(IllegalArgumentException.class, () -> new Matrix(-4, 3));
    }

    @Test
    void throwsExceptionInvalidCols() {
        assertThrows(IllegalArgumentException.class, () -> new Matrix(3, -5));
    }

    @Test
    void throwsExceptionInvalidRowsCols() {
        assertThrows(IllegalArgumentException.class, () -> new Matrix(-3, -5));
    }

    @Test
    void throwsExceptionInvalidArrayTooLarge() {
        assertThrows(IllegalArgumentException.class, () -> new Matrix(3, 5, new double[16]));
    }

    @Test
    void throwsExceptionInvalidArrayTooSmall() {
        assertThrows(IllegalArgumentException.class, () -> new Matrix(3, 5, new double[12]));
    }

    @Test
    void testTranspose0() {
        Matrix matrix = new Matrix(3, 4, new double[]{
                1, 2, 3, 4,
                5, 6, 7, 8,
                9, 4, 2, 1});
        Matrix transpose = matrix.transpose();

        assertEquals(4, transpose.getRows());
        assertEquals(3, transpose.getCols());

        assertArrayEquals(new double[]{
                1, 5, 9,
                2, 6, 4,
                3, 7, 2,
                4, 8, 1
        }, transpose.getElements(), DELTA);
    }

    @Test
    void testMatrixToString() {
        Matrix matrix = new Matrix(4, 2, new double[]{
                2, 9,
                4, 3,
                7, 8,
                1, 5
        });
        assertEquals("[2.0, 9.0]\n[4.0, 3.0]\n[7.0, 8.0]\n[1.0, 5.0]\n", matrix.toString());
    }
}
