import ee.ut.cs.gitlab.maikov.matrix.SquareMatrix;
import org.junit.jupiter.api.Test;

import static org.junit.jupiter.api.Assertions.*;

public class SquareMatrixTest {

    @Test
    public void testMatrixIdentity() {
        SquareMatrix matrix = new SquareMatrix(3);
        matrix.identity();

        assertArrayEquals(new double[]{1, 0, 0, 0, 1, 0, 0, 0, 1}, matrix.getElements());
    }

    @Test
    public void testMatrixTriangularAbove() {
        SquareMatrix matrix = new SquareMatrix(4, new double[]{
                9,  0, 0, 0,
                3,  7, 0, 0,
                5,  4,-2, 0,
                1, -3, 2, 5
        });

        assertTrue(matrix.isTriangular());
    }

    @Test
    public void testMatrixTriangularBelow() {
        SquareMatrix matrix = new SquareMatrix(4, new double[]{
                9, 1, 2, 3,
                0, 7, 2, 4,
                0, 0,-2, 8,
                0, 0, 0, 5
        });

        assertTrue(matrix.isTriangular());
    }

    @Test
    public void testMatrixTriangularNot() {
        SquareMatrix matrix = new SquareMatrix(4, new double[]{
                9, 1, 2, 3,
                0, 7, 2, 4,
                0, 0,-2, 8,
                5, 0, 0, 5
        });

        assertFalse(matrix.isTriangular());
    }

    @Test
    public void testMatrixMinor0() {
        SquareMatrix matrix = new SquareMatrix(4, new double[]{
                9, 1, 2, 3,
                0, 7, 2, 4,
                0, 0,-2, 8,
                5, 0, 0, 5
        });

        SquareMatrix minorMatrix = matrix.minorMatrix(0, 0);

        assertEquals(3, minorMatrix.getRows());
        assertEquals(3, minorMatrix.getCols());

        assertArrayEquals(new double[]{
                7, 2, 4,
                0,-2, 8,
                0, 0, 5
        }, minorMatrix.getElements());
    }

    @Test
    public void testMatrixMinor1() {
        SquareMatrix matrix = new SquareMatrix(4, new double[]{
                9, 1, 2, 3,
               -1, 7, 2, 4,
                4, 0, 0, 8,
                5, 0, 1, 5
        });

        SquareMatrix minorMatrix = matrix.minorMatrix(1, 1);

        assertEquals(3, minorMatrix.getRows());
        assertEquals(3, minorMatrix.getCols());

        assertArrayEquals(new double[]{
                9, 2, 3,
                4, 0, 8,
                5, 1, 5
        }, minorMatrix.getElements());
    }

    @Test
    public void testMatrixDeterminant1() {
        SquareMatrix matrix = new SquareMatrix(2, new double[]{
                4, 2,
                1,-3
        });

        double det = matrix.determinant();
        assertEquals(-14, det);
    }

    @Test
    public void testMatrixDeterminant2() {
        SquareMatrix matrix = new SquareMatrix(3, new double[]{
                4, 0, 0,
                1,-3, 0,
                4, 3,-2
        });

        double det = matrix.determinant();
        assertEquals(24, det);
    }

    @Test
    public void testMatrixDeterminant3() {
        SquareMatrix matrix = new SquareMatrix(3, new double[]{
                1, 4, 3,
                0,-3, 2,
                0, 0,-2
        });

        double det = matrix.determinant();
        assertEquals(6, det);
    }

    @Test
    public void testMatrixDeterminant4() {
        SquareMatrix matrix = new SquareMatrix(3, new double[]{
                1, 4, 3,
                0,-3, 2,
                0, 0, 0
        });

        double det = matrix.determinant();
        assertEquals(0, det);
    }

    @Test
    public void testMatrixDeterminant5() {
        SquareMatrix matrix = new SquareMatrix(3, new double[]{
                0, 4, 3,
                0,-3, 2,
                0, 3, 4
        });

        double det = matrix.determinant();
        assertEquals(0, det);
    }

    @Test
    public void testMatrixDeterminant6() {
        SquareMatrix matrix = new SquareMatrix(3, new double[]{
                7, 4, 3,
                7, 4, 3,
                0, 3, 4
        });

        double det = matrix.determinant();
        assertEquals(0, det);
    }

    @Test
    public void testMatrixDeterminant7() {
        SquareMatrix matrix = new SquareMatrix(3, new double[]{
                7, 5, 7,
                7, 4, 7,
                7, 3, 7
        });

        double det = matrix.determinant();
        assertEquals(0, det);
    }

    @Test
    public void testMatrixDeterminant8() {
        SquareMatrix matrix = new SquareMatrix(3, new double[]{
                7, 5,14,
                7, 4,14,
                7, 3,14
        });

        double det = matrix.determinant();
        assertEquals(0, det);
    }

    @Test
    public void testMatrixDiagonalSum() {
        SquareMatrix matrix = new SquareMatrix(3, new double[]{
                1, 4, 3,
                1,-3, 2,
                3, 4,-2
        });

        double sum = matrix.getDiagonalSum();
        assertEquals(-4, sum);
    }


    @Test
    public void testMatrixDiagonalProduct() {
        SquareMatrix matrix = new SquareMatrix(3, new double[]{
                5, 4, 3,
                1,-3, 2,
                3, 4,-2
        });

        double product = matrix.getDiagonalProduct();
        assertEquals(30, product);
    }
}
